package ru.dbotthepony.mc.otm.container

import net.minecraft.world.Container
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ICapabilityProvider
import ru.dbotthepony.mc.otm.core.ifPresentK

class ContainerIterator(private val container: Container) : MutableIterator<ItemStack> {
	private var index = 0

	override fun hasNext(): Boolean {
		return index < container.containerSize
	}

	override fun next(): ItemStack {
		if (index >= container.containerSize) {
			throw IllegalStateException("Already finished iterating")
		}

		return container[index++]
	}

	override fun remove() {
		if (index == 0) {
			throw IllegalStateException("Never called next()")
		}

		container[index - 1] = ItemStack.EMPTY
	}
}

open class NonEmptyItemStackIterator(protected open val parent: Iterator<ItemStack>) : Iterator<ItemStack> {
	private var itemStack: ItemStack? = null
	private var searched = false

	private fun search() {
		searched = true

		if (itemStack != null) {
			return
		}

		for (stack in parent) {
			if (!stack.isEmpty) {
				itemStack = stack
				return
			}
		}
	}

	override fun hasNext(): Boolean {
		if (!searched) {
			search()
		}

		return itemStack != null
	}

	override fun next(): ItemStack {
		if (!searched) {
			search()
		}

		val itemStack = itemStack ?: throw IllegalStateException("No next element")
		this.itemStack = null
		this.searched = false
		return itemStack
	}
}

class NonEmptyMutableItemStackIterator(override val parent: MutableIterator<ItemStack>) : NonEmptyItemStackIterator(parent), MutableIterator<ItemStack> {
	override fun remove() {
		parent.remove()
	}
}

open class CapabilityIterator<P : ICapabilityProvider, T>(protected open val parent: Iterator<P>, private val cap: Capability<T>) : Iterator<Pair<P, T>> {
	private var provider: P? = null
	private var capability: T? = null
	private var searched = false

	private fun search() {
		searched = true

		if (provider != null) {
			return
		}

		for (provider in parent) {
			provider.getCapability(cap).ifPresentK {
				this.provider = provider
				capability = it
				return
			}
		}
	}

	override fun hasNext(): Boolean {
		if (!searched) {
			search()
		}

		return provider != null
	}

	override fun next(): Pair<P, T> {
		if (!searched) {
			search()
		}

		val provider = provider ?: throw IllegalStateException("No next element")
		val capability = capability ?: throw IllegalStateException("No next element")
		this.provider = null
		this.capability = null
		this.searched = false
		return provider to capability
	}
}

class MutableCapabilityIterator<P : ICapabilityProvider, T>(override val parent: MutableIterator<P>, cap: Capability<T>) : CapabilityIterator<P, T>(parent, cap), MutableIterator<Pair<P, T>> {
	override fun remove() {
		parent.remove()
	}
}

fun Container.iterator() = ContainerIterator(this)
fun Iterator<ItemStack>.nonEmpty() = NonEmptyItemStackIterator(this)
fun MutableIterator<ItemStack>.nonEmpty() = NonEmptyMutableItemStackIterator(this)
fun Container.nonEmptyIterator() = iterator().nonEmpty()

fun <T> Container.iterator(cap: Capability<T>) = CapabilityIterator(nonEmptyIterator(), cap)

fun <P : ICapabilityProvider, T> Iterator<P>.filtered(cap: Capability<T>) = CapabilityIterator(this, cap)
fun <P : ICapabilityProvider, T> MutableIterator<P>.filtered(cap: Capability<T>) = MutableCapabilityIterator(this, cap)

inline fun <T> Container.forEach(cap: Capability<T>, consumer: (Pair<ItemStack, T>) -> Unit) {
	for (value in iterator(cap)) {
		consumer(value)
	}
}

inline fun <T> Container.forEach(cap: Capability<T>, consumer: (ItemStack, T) -> Unit) {
	for ((a, b) in iterator(cap)) {
		consumer(a, b)
	}
}

inline fun <T> Container.forEachItem(cap: Capability<T>, consumer: (ItemStack) -> Unit) {
	for (pair in iterator(cap)) {
		consumer(pair.first)
	}
}

inline fun <T> Container.forEachCapability(cap: Capability<T>, consumer: (T) -> Unit) {
	for (pair in iterator(cap)) {
		consumer(pair.second)
	}
}

inline fun Container.forEach(lambda: (ItemStack) -> Unit) {
	for (value in iterator()) {
		lambda(value)
	}
}

inline fun Container.forEachNonEmpty(lambda: (ItemStack) -> Unit) {
	for (value in nonEmptyIterator()) {
		lambda(value)
	}
}
